/*
 * Kernel-friendly wrappers for many public ifp function.
 * $Id: ifp_dentry.c,v 1.4 2004/09/30 00:24:27 oakhamg Exp $
 *
 * Copyright (C) Geoff Oakham, 2004; <oakhamg@users.sourceforge.net>
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/highmem.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/statfs.h>
#include <linux/smp_lock.h>
#include <linux/backing-dev.h>
#include <linux/usb.h>
#include <linux/nls.h>
#include <linux/writeback.h>
#include <linux/mpage.h>
#include <linux/ctype.h>

#include <asm/uaccess.h>

#include "iriverfs.h"
#include "ifp.h"
#include "ifp_os.h"
#include "ifp_prim.h"

#if defined(__KERNEL__)

int make_path_string(struct dentry * target, struct dentry * root, char separator,
		char * out, int maxlen)
{
	if (target == root) {
		out[0] = separator;
		out[1] = '\0';
		return sizeof(separator);
	} else {
		struct dentry * d;
		char * pos;
		int len = 0;

		for (d = target; d && d != root; d = d->d_parent) {
			//for the path separator
			len += sizeof(separator);
			len += d->d_name.len;
		}
		if (len + 1 > maxlen) {
			printk("make_path_string: buffer (%d) too small for path (%d)\n", maxlen, len);
			return -1;
		}

		pos = out + len;
		*pos = '\0';
		for (d = target; d && d != root; d = d->d_parent) {
			pos -= d->d_name.len;
			memcpy(pos, d->d_name.name, d->d_name.len);

			pos -= sizeof(separator);
			*pos = separator;
		}

		BUG_ON(pos != out);
		return len;
	}
}

//The convience methods..
static inline int call_dentry_as_utf8(
	struct ifp_device * dev, struct dentry * d,
	int(* fn)(struct ifp_device * dev, char *, int n)
	)
{
	//u8 buf[IFP_BUFFER_SIZE];
	u8 * buf = get_iriver_buffer();
	int i, tsize;

	tsize = make_path_string(d, d->d_sb->s_root, '\\', buf, IFP_BUFFER_SIZE);
	if (tsize < 0) {
		printk("[call_dentry_as_utf8]: error creating pathname for %ld, err %d\n",
			d->d_inode->i_ino, tsize);
		i = tsize;
		goto out;
	}
	i = fn(dev, buf, IFP_BUFFER_SIZE);
	if (i < 0) {
		make_path_string(d, d->d_sb->s_root, '/', buf, IFP_BUFFER_SIZE);
		printk("[call_dentry_as_utf8] fn failed on ifp:/%s, err %d\n", buf, i);
	}

out:
	put_iriver_buffer(buf); buf = NULL;
	return i;
}

int ifp_file_open_new_d(struct ifp_device * dev, struct dentry * d, int filesize)
{
	//u8 buf[IFP_BUFFER_SIZE];
	u8 * buf = get_iriver_buffer();
	int i, tsize;

	tsize = make_path_string(d, d->d_sb->s_root, '\\', buf, IFP_BUFFER_SIZE);
	if (tsize < 0) {
		printk("[call_dentry_as_utf8]: error creating pathname for %ld, err %d\n",
			d->d_inode->i_ino, tsize);
		i = tsize;
		goto out;
	}

	i = ifp_file_open_new(dev, buf, IFP_BUFFER_SIZE, filesize);
	if (i) {
		make_path_string(d, d->d_sb->s_root, '/', buf, IFP_BUFFER_SIZE);
		printk("[ifp_file_open_new_d] failed to create 'ifp:/%s', err %d\n", buf, i);
	}
out:
	put_iriver_buffer(buf); buf = NULL;
	return i;
}

int ifp_file_open_d(struct ifp_device * dev, struct dentry * d)
{ return call_dentry_as_utf8(dev, d, ifp_file_open); }

int ifp_dir_open_d(struct ifp_device * dev, struct dentry * d)
{ return call_dentry_as_utf8(dev, d, ifp_dir_open); }

int ifp_delete_d(struct ifp_device * dev, struct dentry * d)
{ return call_dentry_as_utf8(dev, d, ifp_delete); }

int ifp_mkdir_d(struct ifp_device * dev, struct dentry * d)
{ return call_dentry_as_utf8(dev, d, ifp_mkdir); }

int ifp_rmdir_d(struct ifp_device * dev, struct dentry * d)
{ return call_dentry_as_utf8(dev, d, ifp_rmdir); }

int ifp_read_open_d(struct ifp_device * dev, struct dentry * d)
{ return call_dentry_as_utf8(dev, d, ifp_read_open); }

int ifp_rename_d(struct ifp_device * dev,
        struct dentry * old, struct dentry * new)
{
        int i;
        struct super_block * sb = old->d_sb;
        struct dentry * root = sb->s_root;

        char * old_file = get_iriver_buffer();
        char * new_file = get_iriver_buffer();

        i = make_path_string(old          , root, '\\', old_file,
                        IFP_BUFFER_SIZE);
        if (i<0){printk("[rename_d] old file string, err %d.\n", i); goto out;}
        i = make_path_string(new          , root, '\\', new_file,
                        IFP_BUFFER_SIZE);
        if (i<0){printk("[rename_d] new file string, err %d.\n", i); goto out;}

	if (S_ISDIR(old->d_inode->i_mode)) {
		i = ifp_rename_dir(dev, old_file, new_file, IFP_BULK_MAXPATHLEN);
	} else {
		i = ifp_rename_file(dev, old_file, new_file, IFP_BULK_MAXPATHLEN);
	}

out:
        put_iriver_buffer(old_file);
        put_iriver_buffer(new_file);
        return i;
}

int ifp_list_dirs_d(
        struct ifp_device * dev,
        struct dentry * dentry,
        int(*callbk)(void *, int, char *, int, int),
        void * context)
{
        int i = 0;
        u8 * dir = get_iriver_buffer();

        i = make_path_string(dentry, dentry->d_sb->s_root, '\\', dir,
                IFP_BULK_MAXPATHLEN);
        if (i < 0) {
                printk("[ifp_list_dirs_d] mkpath returned %d.\n", i);
                goto out;
        }

        i = ifp_list_dirs(dev, dir, IFP_BULK_MAXPATHLEN, callbk, context);
        if (i) {
                printk("[ifp_list_dirs_d] ifp_list_dirs failed with err %d.\n", i);
                goto out;
        }

out:
        put_iriver_buffer(dir); dir = NULL;
        return i;
}

int ifp_count_subdirs_d(struct ifp_device * dev, struct dentry * dentry)
{
        int i = 0;
        u8 * s = get_iriver_buffer();
        i = make_path_string(dentry, dentry->d_sb->s_root, '\\', s,
                IFP_BULK_MAXPATHLEN);
        if (i < 0) {
                printk("[ifp_list_dirs_d] mkpath returned %d.\n", i);
                goto out;
        }
        i = ifp_count_subdirs(dev, s, IFP_BULK_MAXPATHLEN);
        if (i < 0) {
                ifp_err_i(i, "ifp_count_subdirs failed.");
        }
out:
        return i;
}

int ifp_write_open_d(struct ifp_device * dev, struct dentry * d, int filesize)
{
	int i = 0;
	struct dentry * root = NULL;
	char * file  = get_iriver_buffer();

	root = d->d_sb->s_root;
	i = make_path_string(d, root, '\\', file, IFP_BUFFER_SIZE);
	if (i<0){printk("[ifp_write_open_d] file string, err %d.\n", i); goto out;}
	i = ifp_write_open(dev, file, IFP_BUFFER_SIZE, filesize);
	if (i) {
		printk("[ifp_write_open_d] ifp_write_open failed, err %d\n", i);
		goto out;
	}

out:
	put_iriver_buffer(file); file = NULL;
	return i;
}

#endif // __KERNEL__

