/*
 * IRiver ifp supporting functions, a basic (primative) API.
 * $Id: ifp_prim.c,v 1.20.2.1 2004/10/23 16:36:10 oakhamg Exp $
 *
 * Copyright (C) Geoff Oakham, 2004; <oakhamg@users.sourceforge.net>
 */

#define IFP_ENOENT 2
#define IFP_EIO    5

#include "ifp.h"
#include "ifp_os.h"
#include "ifp_prim.h"

//Constants needed for autodetection.  (Platform-indendent, I might add.)

//This list is a guestimate.  I have good reason to believe it works for
//iFP-1xx through iFP-8xx though.
int ifp_product_ids[IFP_PRODUCT_IDs] = {
	0x1001,
	0x1003,
	0x1005,
	0x1007,
	0x1008,
	0x1010
};
char * ifp_product_strings[IFP_PRODUCT_IDs] = {
	"iFP-1xx",
	"iFP-3xx",
	"iFP-5xx",
	"iFP-7xx",
	"iFP-8xx",
	"iFP-10xx"
};

#ifdef IFP_DEBUG_USB_SNOOPING
static char const * ifp_human_commands[] = {
	"ping", //0x00
	"0x01", //0x01
	"02_command", //0x02
	"0x03", //0x03
	"0x04", //0x04
	"f.open", //0x05
	"f.new", //0x06
	"download", //0x07
	"upload", //0x08
	"0x09", //0x09
	"0x0a", //0x0a
	"f.size", //0x0b
	"0x0c", //0x0c
	"f.close", //0x0d
	"delete", //0x0e
	"d.open", //0x0f
	"d.next", //0x10
	"d.close", //0x11
	"mkdir", //0x12
	"rmdir", //0x13
	"capacity", //0x14
	"availa", //0x15
	"format", //0x16
	"firmware update", //0x17
	"buffsz", //0x18
	"0x19", //0x19
	"getFAT page", //0x1a
	"d.next_debug", //0x1b
	"setFAT page", //0x1c
	"getRadio stations", //0x1d
	"setRadio stations", //0x1e
	"0x1f", //0x1f
};

static char const * ifp_human_02_commands[] = {
	"string", //0x00
	"02_01", //0x01
	"02_02", //0x02
	"firmwr", //0x03
	"delta", //0x04
	"02_05", //0x05
	"flush", //0x06
	"02_07", //0x07
	"battery", //0x08
};

static int _ifp_control_debug(struct ifp_device * dev,
	int command, int arg1, int arg2, int r1, int * pr2,
	int ivalue
	)
{
	int i = 0;
	int r2 = -1;
	char const * s_command = NULL;


	if (pr2) {
		r2 = *pr2;
	}

	if (command == 0x02 && arg2 >= 0 && arg2 <= 0x08) {
		s_command = ifp_human_02_commands[arg2];
	} else if (command >= 0 && command <= 0x1f) {
		s_command = ifp_human_commands[command];
	} else {
		s_command = "";
	}

	ifp_print("ifp: %-6.6s c0%02x %04x %04x | %04x",
		s_command, command, arg1, arg2, r1);

	if (pr2) {
		ifp_print(" %04x", r2);
	}

	if (ivalue) {
		ifp_print(" i=%d", ivalue);
	}

	ifp_print("\n");
	return i;
}

static inline int ifp_control_debug_info(struct ifp_device * dev,
	int command, int arg1, int arg2, int r1, int * pr2,
	int ivalue
	)
{
	return _ifp_control_debug(dev, command, arg1, arg2, r1, pr2, ivalue);
}
#else //IFP_DEBUG_USB_SNOOPING
static inline int ifp_control_debug_info(struct ifp_device * dev,
	int command, int arg1, int arg2, int r1, int * pr2, int ivalue)
{
	return 0;
}
#endif //IFP_DEBUG_USB_SNOOPING

int ifp_control_send(struct ifp_device * dev, int command,
	int arg1, int arg2, int * pr2)
{
	int i, r1;
	i = ifp_os_control_send(dev, command, arg1, arg2, &r1, pr2);
	ifp_control_debug_info(dev, command, arg1, arg2, r1, pr2, i);
	if (i < 0) {
		ifp_err_i(i, "error sending control value");
	} else if (i) {
		//caller allready warned about unexpected value.
		i = 0;
	}

	if (i == 0) {
		i = r1;
	}

	return i;
}

//Convience method to convert ifp's "success" return value to 0
int ifp_control_send_bool(struct ifp_device * dev, int command, int arg1, int arg2,
	int * ret)
{
	int i = ifp_control_send(dev, command, arg1, arg2, ret);
	if (i == 0) {
		//FIXME: change to i = 1;
		//i = -1;
		i = 1;
	} else if (i == 1) {
		i = 0;
	}
	return i;
}


#ifdef IFP_DEBUG_USB_RAWDATA
static int hexdump_line(void * pp, int n, int lw)
{
	ifp_u8 * p = pp;
	int j, c;

	for (j=0; j!=lw; j++) {
		if (j >= n) {
			printk("  ");
		} else {
			printk("%02x", (int)(p[j]));
		}
		if (j % 4 == 3) {
			printk(" ");
		}
	}

	for (j=0; j!=lw; j++) {
		if (j >= n) {
			break;
			//printk(" ");
		} else {
			c = (int)p[j];
			if (isprint(c)) {
				printk("%c", c);
			} else {
				printk(".");
			}
		}
	}

	return 0;
}

static int hexdump(char * prefix, void * pp, int n)
{
	ifp_u8 * p = pp;
	int const lw = 32;
	int zeros = 0;
	int j = n;
	int total = n;

	while (p[j-1] == 0) {
		j--;
	}
	if (n-j > 3) {
		zeros = n - j;
		n = j;
	}
	
	while (n > 0) {
		printk("%s", prefix);
		hexdump_line(p, n, lw);
		n -= lw;
		if (n > 0) {
			p += lw;
		} else {
			if (zeros > 0) {
				printk(" [%d]", total);
			}
		}
		printk("\n");
	}
	return 0;
}
#endif


//
// primatives
//

//Returns a negative number on error, 0 on success, and the number of
//bytes written (if the whole block wasn't written).
static inline int _ifp_push(struct ifp_device * dev, void * p, int n)
{
	int i = 0;

	i = ifp_os_push(dev, p, n);

#ifdef IFP_DEBUG_USB_RAWDATA
	hexdump("<--dn  ", p, n);
#endif

	return i;
}

/* Note that 'n' must be a power of 2. */
static inline int _ifp_pop(struct ifp_device * dev, void * p, int n)
{
	int i = 0;

	i = ifp_os_pop(dev, p, n);

#ifdef IFP_DEBUG_USB_RAWDATA
	hexdump("-->up  ", p, n);
#endif

	return i;
}

static inline int _ifp_push_unicode(struct ifp_device * dev, char * s, int n,
		int blocksize)
{
	int i;
	ifp_u8 * buf = dev->b1;

	memset(buf, 0, IFP_BUFFER_SIZE);

	//Warning about possible buffer overrun: here 'n' is the
	//size of the buffer 's', but we have no way of preventing
	//'buf' from being overrun.  (We can tell afterwards, but by
	//then it's too late.)
	
	i = ifp_utf8_to_utf16((wchar_t *)buf, IFP_BUFFER_SIZE/sizeof(wchar_t), s, n);
	//i = utf8_mbstowcs((wchar_t *) buf, s, n);
	//BUG_ON(i > IFP_BUFFER_SIZE);
	if (i < 0) {
		ifp_err_i(i, "error conversion from utf8 to utf16.  n=%d.\n", n);
		return i;
	}
	i = ifp_os_push(dev, buf, blocksize);

#ifdef IFP_DEBUG_USB_SNOOPING
	printk("<--dn '%s'[unicode]\n", s);
#endif // IFP_DEBUG_USB_SNOOPING
	return i;
}

static inline int _ifp_pop_unicode(struct ifp_device * dev, char * s, int n,
		int blocksize)
{
	int i;
	ifp_u8 * buf = dev->b1;

	memset(buf, 0, IFP_BUFFER_SIZE);
	i = ifp_os_pop(dev, buf, blocksize);
	if (i) {
		//Either an error occured, or fewer bytes than 'n' was read.
		if (i < 0) {
			ifp_err_i(i, "pop error.");
		} else {
			ifp_err("pop read only %d bytes.", i);
		}
		return i;
	}
	i = ifp_utf16_to_utf8(s, n, (wchar_t *)buf, IFP_BUFFER_SIZE/sizeof(wchar_t));
	//i = utf8_wcstombs(s, (wchar_t *)buf, n);
	//BUG_ON(i >= n);
	((char *)s)[i] = '\0';
#ifdef IFP_DEBUG_USB_SNOOPING
	printk("-->up '%s'[unicode]\n", s);
#endif // IFP_DEBUG_USB_SNOOPING

	return 0;
}

#if 0
//Returns a negative number on error, 0 on success, and the number of
//bytes written (if the whole block wasn't written).
static inline int _ifp_set_buffer_size(struct ifp_device * dev, int n,
	int force)
{
	int i = 0, allowed;
	if (dev->last_buffer_size != n || force) {
		i = ifp_control_send(dev, IFP_SET_BUFFER, n, 0, &allowed);
		if (allowed >= 0) {
			dev->last_buffer_size = allowed;
		}
		if (i == 1 && allowed == n) {
			return 0;
		}
	}
	//return 0;
	return i ? i : 0;
}
#endif

static inline int _ifp_file_open(struct ifp_device * dev, int type)
{
	int i = ifp_control_send(dev, IFP_FILE_OPEN, type, 0, NULL);
	if (i == 0) {
		//doesn't exist
		i = 1;
	} else if (i == 1) {
		//exists, ok
		i = 0;
	} else if (i > 1) {
		//wtf
		ifp_err_i(i, "unexpected return code for f.open.");
		i = -1;
	}
	return i;
}

static inline int _ifp_file_open_new(struct ifp_device * dev, int size)
{ return ifp_control_send_bool(dev, IFP_FILE_OPEN_NEW, size, 0, NULL); }

static inline int _ifp_file_close(struct ifp_device * dev)
{ return ifp_control_send_bool(dev, IFP_FILE_CLOSE, 0, 0, NULL); }

static inline int _ifp_file_size(struct ifp_device * dev)
{
	int i, n;
	i = ifp_control_send_bool(dev, IFP_FILE_SIZE, 0, 0, &n);
	if (i) {
		return i;
	}
	return n;
}

#if 0
/* returns 0 on success, 1 on "device not ready!" and negative on error. */
static inline int _ifp_file_download(struct ifp_device * dev, int bytes, int * actual)
{
	int i;
	i = ifp_control_send(dev, IFP_FILE_DOWNLOAD, bytes, 0, NULL);
	if (actual) {
		*actual = i;
	}
	if (i < 0) {
		ifp_err_i(i, "error sending code to download block.");
		return i;
	} else if (i == bytes) {
		return 0;
	} else if (i > bytes) {
		ifp_err_i(i, "Something's wierd.  The return value is larger than %d", bytes);
		return -1;
	} else if (i == 0) {
		ifp_wrn("warning: return value is 0 instead of %d, which is often a sign of corruption.", bytes);
		return 1;
	} else {
		//ifp_wrn("[_ifp_file_download] warning: got %d instead of %d.\n", i, bytes);
		return 0;
	}
}
#endif

static inline int _ifp_dir_open(struct ifp_device * dev)
{ return ifp_control_send(dev, IFP_LS_OPEN, 0, 0, NULL); }

static inline int _ifp_dir_next(struct ifp_device * dev, int mode)
{ return ifp_control_send(dev, IFP_LS_NEXT, mode, 0, NULL); }

static inline int _ifp_dir_close(struct ifp_device * dev)
{ return ifp_control_send_bool(dev, IFP_LS_CLOSE, 0, 0, NULL); }


//Returns a negative number on error, 0 on success, and the number of
//bytes written (if the whole block wasn't written).
static inline int _ifp_file_upload(struct ifp_device * dev, int bytes)
{
	int i;
	i = ifp_control_send(dev, IFP_FILE_UPLOAD, bytes, 0, NULL);
	if (i >= 0 && i == bytes) {
		return 0;
	}
	return i < 0 ? i : -1;
}

static inline int _ifp_file_flush(struct ifp_device * dev)
{
	int n;
	return ifp_control_send_bool(dev, IFP_02_COMMAND, 0, IFP_02_UPLOAD_FLUSH, &n);
	//printk("[_ifp_file_flush] info, n=%d\n", n);
}

static inline int _ifp_delete(struct ifp_device * dev)
{ return ifp_control_send_bool(dev, IFP_FILE_DELETE, 0, 0, NULL); }

static inline int _ifp_mkdir(struct ifp_device * dev)
{ return ifp_control_send_bool(dev, IFP_DIR_CREATE, 0, 0, NULL); }

static inline int _ifp_rmdir(struct ifp_device * dev)
{ return ifp_control_send_bool(dev, IFP_DIR_DELETE, 0, 0, NULL); }

//return 1 on success, 0 on 'done', and negative on error
static inline int _ifp_dir_next_debug(struct ifp_device * dev, int mode)
{ return ifp_control_send(dev, IFP_LS_NEXT_DEBUG, mode, 0, NULL); }

static inline int _ifp_get_fat_page(struct ifp_device * dev, int dir, int page)
{ return ifp_control_send_bool(dev, IFP_GET_FAT_PAGE, dir, page, NULL); }

static inline int _ifp_set_fat_page(struct ifp_device * dev)
{ return ifp_control_send_bool(dev, IFP_SET_FAT_PAGE, 0, 0, NULL); }

static inline int get_control_value(struct ifp_device * dev,
	int cmd, int arg1, int arg2)
{
	int i,n;
	i = ifp_control_send_bool(dev, cmd, arg1, arg2, &n);
	if (i < 0) {
		return i;
	}
	return n;
}

static inline int _ifp_ping(struct ifp_device * dev)
{ return get_control_value(dev, IFP_PING, 0, 0); }

static inline int _ifp_battery(struct ifp_device * dev)
{ return get_control_value(dev, IFP_02_COMMAND, 0, IFP_02_BATTERY); }

static inline int _ifp_capacity(struct ifp_device * dev)
{ return get_control_value(dev, IFP_GET_CAPACITY, 0, 0); }

static inline int _ifp_freespace(struct ifp_device * dev)
{ return get_control_value(dev, IFP_GET_FREE, 0, 0); }

static inline int _ifp_model(struct ifp_device * dev)
{ return get_control_value(dev, IFP_02_COMMAND, 0, IFP_02_STRING); }

static inline int _ifp_firmware_version(struct ifp_device * dev)
{ return get_control_value(dev, IFP_02_COMMAND, 0, IFP_02_FIRMWARE); }

static inline int _ifp_delta(struct ifp_device * dev)
{ return get_control_value(dev, IFP_02_COMMAND, 0, IFP_02_DELTA); }

//
//public functions (to this module)
//

int ifp_init(struct ifp_device * dev, void * device_handle) {
	int i = 0;

	if (device_handle == NULL) {
		ifp_err("device handle is NULL.");
		i = -1;
		goto out;
	}

	i = ifp_os_init(dev, device_handle);
	ifp_err_jump(i, out, "ifp_os_init error.");

	dev->last_buffer_size = 0;
	dev->mode = IFP_MODE_NONE;

	i = ifp_selftest(dev);
	ifp_err_jump(i, out_err, "self test failed.");

	return 0;

out_err:
	ifp_os_finalize(dev);
out:
	return i;
}

int ifp_finalize(struct ifp_device * dev) {
	int i = 0;
	
	i = ifp_os_finalize(dev);
	ifp_err_jump(i, out, "ifp_os_finalize returned an error.");

out:
	return i;
}

int ifp_file_open(struct ifp_device * dev, char * s, int n)
{
	int i;
	//i = _ifp_set_buffer_size(dev, n, 1);
	i = _ifp_set_buffer_size(dev, IFP_PATH_XFER_SIZE, 1);
	ifp_err_jump(i, out, "set buffer failed");
	i = _ifp_push_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	ifp_err_jump(i, out, "push failed");
	i = _ifp_file_open(dev, IFP_FILE);
	if (i != 0 && i != 1) { ifp_err_i(i, "open directive failed"); return i; }

out:
	return i;
}

int ifp_file_open_new(struct ifp_device * dev, char * s, int n, int filesize)
{
	int i;
	//i = _ifp_set_buffer_size(dev, n, 1);
	i = _ifp_set_buffer_size(dev, IFP_PATH_XFER_SIZE, 0);
	if (i) { ifp_err_i(i, "set buffer failed"); return i; }
	i = _ifp_push_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	if (i) { ifp_err_i(i, "push failed"); return i; }
	i = _ifp_file_open_new(dev, filesize);
	if (i) { ifp_err_i(i, "open_new directive failed"); return i; }

	return i;
}

int ifp_file_close(struct ifp_device * dev)
{
	return _ifp_file_close(dev);
}

int ifp_file_size(struct ifp_device * dev)
{
	return _ifp_file_size(dev);
}

//I'd like to move this out of this file, eventually
int ifp_read_seek_forward(struct ifp_device * dev, int count, int blocksize)
{
	int i, j, actual;
	i = _ifp_set_buffer_size(dev, blocksize, 1);
	if (i) {
		ifp_err_i(i, "set buffer failed");
		return i > 0 ? -IFP_EIO : i;
	}
	for (j=0; j != count; j++) {
		i = _ifp_file_download(dev, blocksize, &actual);
		if (i) {
			ifp_err_i(i, "download control message failed");
			return i;
		}
		if (actual != blocksize) {
			i = -IFP_EIO;
			ifp_err("seek failed,  I can't handle "
				"getting %d bytes instead of %d\n", i, blocksize
			);
			return -IFP_EIO;
		}
	}
	return 0;
}


//returns the number of bytes received, or a negative number on error.
int ifp_file_download(struct ifp_device * dev, void * p, int n)
{
	int i;
	int actual;

	if (n == 0) {
		ifp_err("refusing to download 0 bytes.");
		return -1;
	}

	i = _ifp_set_buffer_size(dev, n, 1);
	if (i) {
		ifp_err_i(i, "set buffer failed");
		return i > 0 ? -IFP_EIO : i;
	}
	i = _ifp_file_download(dev, n, &actual);
	if (i) {
		ifp_err_i(i, "download control code failed");
		return i;
	}

	if (actual == 0) {
		ifp_wrn("warning: zero bytes available for download.");
		return 0;
	}
	//printk("[ifp_file_download] download returned %d\n",i);
	n = actual;

	i = _ifp_pop(dev, p, n);
	if (i) {
		//Either an error occured, or fewer bytes than 'n' was read.
		if (i < 0) {
			ifp_err_i(i, "pop failed");
		} else {
			ifp_err_i(i, "pop returned an unexpected value (fewer bytes read than expected.)");
			i = -1;
		}
		return i;
	}

	return n;
}

int ifp_file_upload(struct ifp_device * dev, void * p, int n)
{
	int i;
	if (n == 0) {
		ifp_err("refusing to upload 0 bytes.");
		return -1;
	}
	i = _ifp_set_buffer_size(dev, n, 1);
	ifp_err_jump(i, out, "set buffer failed");
	i = _ifp_push(dev, p, n);
	ifp_err_jump(i, out, "push failed");
	i = _ifp_file_upload(dev, n);
	ifp_err_jump(i, out, "upload control code failed");

out:
	return i;
}

int ifp_file_flush(struct ifp_device * dev)
{
	unsigned char buf[4];
	int i;
	int status;
	int n = 0;
	int const obsene_limit = 1000;

	do {
		//Sleep even on the first loop.. it's good for soul.. call it "kerma"
		i = ifp_os_sleep(100);
		ifp_err_jump(i, out, "trouble falling asleep, loading Dickens..");

		i = _ifp_file_flush(dev);
		ifp_err_jump(i, out, "error issuing 'flush' command");

		i = _ifp_pop(dev, buf, sizeof(buf));
		ifp_err_jump(i, out, "pop failed");

		//flush returns 1 when it's ready to continue, but 2 when
		//it's still writing to flash.
		status = ifp_os_le32_to_cpup(buf);
		if (status == 1) {
			//done
			return 0;
		}
		if (n == 5) {
			ifp_wrn("[ifp_file_flush] waiting for device 'ready'.");
			ifp_wrn("[ifp_file_flush] (normally it doesn't take this long).");
		}

		//infinite loop plug
		n++;
		if (n == obsene_limit) {
			ifp_err("timeout: hit obscene limit");
			i = -1; goto out;
		}
	} while(status == 2);

	ifp_err("unrecognized return value %d.",i);
	i = -1;
out:
	return i;
}

int ifp_ping(struct ifp_device * dev)
{
	return _ifp_ping(dev);
}

//Returns the battery's charge as an integer--typcially between 0 and 4.
int ifp_battery(struct ifp_device * dev)
{
	ifp_u8 buf[4];
	int n, i = 0;
	n = _ifp_battery(dev);
	if (n < 0) {
		ifp_err_i(n, "error reading battery");
		return n;
	}
	IFP_BUG_ON(n != sizeof(buf));
	i = _ifp_pop(dev, buf, sizeof(buf));
	ifp_err_jump(i, out, "pop failed");

	n = ifp_os_le32_to_cpup(buf);

	return n;
out:
	return i;
}

int ifp_dir_open(struct ifp_device * dev, char * s, int n)
{
	int i;
	//i = _ifp_set_buffer_size(dev, IFP_BUFFER_SIZE, 1);
	i = _ifp_set_buffer_size(dev, IFP_PATH_XFER_SIZE, 1);
	ifp_err_jump(i, out, "set buffer failed");
	i = _ifp_push_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	ifp_err_jump(i, out, "push failed");
	i = _ifp_dir_open(dev);
	if (i < 0) {
		ifp_err_i(i, "open directive failed");
	} else if (i == 0) {
		//dir doesn't exist, so it wasn't opened.  This isn't
		//an error, so don't print any messages to the user.
		i = 1;
	} else if (i == 1) {
		//dir found.. open succeeded.
		i = 0;
	} else {
		ifp_err_i(i, "unexpected return value");
		i = -1;
	}

out:
	return i;
}

int ifp_dir_next(struct ifp_device * dev, void * s, int n, int mode)
{
	int i;
	//printk("[ifp_dir_next] buffer is %#03x, n=%#03x.\n", sizeof(buf), n);
	//i = _ifp_set_buffer_size(dev, IFP_BUFFER_SIZE, 1);
	
	//FIXME: IFP_MAXLSLEN   0x80
	//use it for 'n' in here.  Or check that n is 0x80
	i = _ifp_dir_next(dev, mode);
	if (i < 0) {
		ifp_err_i(i, "error requesting next filename");
		return i;
	} else if (i == 0) {
		//no more files to return.  Just to be nice,
		//we're putting a null terminator on s.
		if (n > 0) {
			*((char *)s) = 0;
		}
		return i;
	}
	mode = i;

	i = _ifp_pop_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	if (i) {
		ifp_err_i(i, "pop failed");
		return i > 0 ? -1 : i;
	}
	return mode;
}

int ifp_dir_close(struct ifp_device * dev)
{
	return _ifp_dir_close(dev);
}

int ifp_delete(struct ifp_device * dev, char * s, int n)
{
	int i = 0;
	i = _ifp_set_buffer_size(dev, IFP_PATH_XFER_SIZE, 1);
	ifp_err_jump(i, out, "set buffer failed");
	i = _ifp_push_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	ifp_err_jump(i, out, "push failed");
	i = _ifp_delete(dev);
	if (i < 0) {
		ifp_err_i(i, "open request failed");
		return i;
	} else if (i == 1) {
		ifp_err("file not found");
		return -IFP_ENOENT;
	}
	i = 0;

out:
	return i;
}

int ifp_mkdir(struct ifp_device * dev, char * s, int n)
{
	int i;
	i = _ifp_set_buffer_size(dev, IFP_PATH_XFER_SIZE, 1);
	ifp_err_jump(i, out, "set buffer failed");
	i = _ifp_push_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	ifp_err_jump(i, out, "push failed");
	i = _ifp_mkdir(dev);
	if (i < 0) {
		ifp_err_i(i, "open failed");
		return i;
	} else if (i == 1) {
		ifp_err("no such directory");
		return -IFP_ENOENT;
	}

	i = 0;
out:
	return i;
}

int ifp_rmdir(struct ifp_device * dev, char * s, int n)
{
	int i;
	i = _ifp_set_buffer_size(dev, IFP_PATH_XFER_SIZE, 1);
	ifp_err_jump(i, out, "set buffer failed");
	i = _ifp_push_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	ifp_err_jump(i, out, "push failed");
	i = _ifp_rmdir(dev);
	if (i < 0) {
		ifp_err_i(i, "open failed");
		return i;
	} else if (i == 1) {
		ifp_err("no such directory");
		return -IFP_ENOENT;
	}

	i = 0;
out:
	return i;
}

int ifp_capacity(struct ifp_device * dev)
{ return _ifp_capacity(dev); }
int ifp_freespace(struct ifp_device * dev)
{ return _ifp_freespace(dev); }

int ifp_delta(struct ifp_device * dev, int * values)
{
	ifp_u8 buf[8];
	int i = 0;
	int bytes;
	int nn;

	bytes = _ifp_delta(dev);
	if (bytes < 0) {
		i = bytes;
		ifp_err_jump(i, out, "error sending control code");
	} else if (bytes > 8) {
		i = -1;
		ifp_err("unexpected buffer size of %d instead of %d",
			bytes, (int)sizeof(buf));
		goto out;
	}
	if (bytes != 8 && bytes != 4) {
		ifp_wrn("interesting, %d bytes are being sent", bytes);

	}

	i = _ifp_pop(dev, buf, bytes);
	ifp_err_jump(i, out, "pop failed");

	values[0] = buf[0];
	values[1] = buf[1];
	values[2] = buf[2];
	values[3] = buf[3];

	if (bytes > 4) {
		nn = ifp_os_le32_to_cpup(buf + 4);
		if ( nn != 0xffffffff ) {
			ifp_wrn("interesting, the last 4 bytes are %08x.", nn);
		}
	}

out:
	return i;
}

int ifp_model(struct ifp_device * dev, char * s, int size)
{
	int n, i = 0;
	n = _ifp_model(dev);
	if (n < 0) {
		ifp_err_i(n, "error reading device model string");
		return n;
	}
	if (n >= size) {
		ifp_wrn("warning: the buffer is too small for the model string.  Truncating.  (%d instead of %d.)", n, size);
		n = size - 1;
	}
	i = _ifp_pop(dev, s, n);
	ifp_err_jump(i, out, "pop failed");

	s[n] = '\0';

out:
	return i;
}

/*Returns a BCD value.*/
int ifp_firmware_version(struct ifp_device * dev)
{
	ifp_u8 * s = dev->b1;
	const int transfer_size = 64;
	int n, i;

	n = _ifp_firmware_version(dev);
	if (n < 0) {
		ifp_err_i(n, "error reading device model string");
		return n;
	}
	//printk("[ifp_firmware_version] info, asking for %d bytes.\n", n);
	if (n >= transfer_size) {
		ifp_wrn("warning: the buffer is too small for the firmware string.  Truncating.  (%d instead of %d.)", n, transfer_size);
		n = transfer_size - 1;
	}
	i = _ifp_pop(dev, s, n);
	if (i) {
		ifp_err_i(i, "pop failed");
		return i >= 0 ? -1 : i;
	}
	//s[n] = '\0';
	
	return ifp_os_le32_to_cpup(s);
}

// 0x0100 >= sizeof(raw) >= 0x0010
int ifp_dir_next_debug(struct ifp_device * dev, char * s, int n, int mode,
	int * dir, int * page, int * offset, int * chipmunk)
{
	ifp_u8 * spec_buff = dev->b1;

	int i;
	if (dev == NULL) {
		ifp_err("dev is NULL..");
	}

	//mode = 0;
	i = _ifp_dir_next_debug(dev, mode);
	if (i == 0) {
		//no more files to return.  Just to be nice,
		//we're putting a null terminator on s.
		if (n > 0) {
			*((char *)s) = 0;
		}
		return 0;
	} else if (i < 0 || i != 1) {
		ifp_err_i(i, "error getting next file");
		return i < 0 ? i : -1;
	}

	i = _ifp_pop_unicode(dev, s, n, IFP_PATH_XFER_SIZE);
	if (i) {
		//Either an error occured, or fewer bytes than 'n' was read.
		if (i < 0) {
			ifp_err_i(i, "pop error");
		} else {
			ifp_err_i(i, "unexpected pop return value");
		}
		return i;
	}

	i = _ifp_dir_next_debug(dev, mode);
	if (i != 1) {
		ifp_err_i(i, "error requesting file debug info for %s", s);
		return i < 0 ? i : -1;
	}
	//buffer is actually IFP_BUFFER_SIZE
	i = _ifp_pop(dev, spec_buff, 0x0100);
	if (i) {
		ifp_err_i(i, "error getting file debug info for %s", s);
		return i < 0 ? i : -1;
	}
	if (dir) {
		*dir = ifp_os_le16_to_cpup(spec_buff);

		if (*dir != ifp_os_le16_to_cpup(spec_buff + 2)) {
			ifp_wrn("warning %d != %d [2]",
				*dir, ifp_os_le16_to_cpup(spec_buff + 2));
		}
		if (*dir != ifp_os_le16_to_cpup(spec_buff + 4)) {
			ifp_wrn("warning %d != %d [4]",
				*dir, ifp_os_le16_to_cpup(spec_buff + 4));
		}
		//BUG_ON(*dir != ifp_os_le16_to_cpup(spec_buff + 2));
		//BUG_ON(*dir != ifp_os_le16_to_cpup(spec_buff + 4));
	}

	if (chipmunk) {
		*chipmunk = ifp_os_le16_to_cpup(spec_buff + 6);
	}

	//Combined with the directory number, this is effectively an 'inode number'.
	i = ifp_os_le16_to_cpup(spec_buff + 8);
	if (i <= 2) {
		//'1' and '2' appear to be reserved for "VOICE" and "RECORD" respectively.
		//I don't think you can manipulate them.  (Not sure why you'd want to anyways.)
		if (page)
			*page = -1;
		if (offset)
			*offset = i;
	} else {
		i -= 2;
		if (page)
			*page = (i)/16;
		if (offset)
			*offset = (i)%16;
	}

	i = ifp_os_le16_to_cpup(spec_buff + 10);
	if (i != 0x0100) {
		ifp_wrn("warning: the field at +10 is %04x instead of 0x0100."
			"  (For %s)", i, s);
	}

	mode = ifp_os_le16_to_cpup(spec_buff + 12);

	if (mode == 0) {
		ifp_err_i(mode, "more error (on %s)", s);
		return -1;
	}

	return mode;
}

int ifp_get_fat_page(struct ifp_device * dev, int dir, int page,
	void * p, int n)
{
	const int DOWNLOAD_PAGE_SIZE = 0x0100;
	int i = 0;

	IFP_BUG_ON(n < DOWNLOAD_PAGE_SIZE*2);
	i = _ifp_get_fat_page(dev, dir, page);
	ifp_err_jump(i, out,
		"error requesting chuck #1 of (%#x, %#x)\n", dir, page);

	i = _ifp_pop(dev, p, DOWNLOAD_PAGE_SIZE);
	ifp_err_jump(i, out,
		"error downloading chuck #1 of (%#x, %#x)\n", dir, page);

	i = _ifp_get_fat_page(dev, dir, page);
	ifp_err_jump(i, out,
		"error requesting chuck #2 of (%#x, %#x)\n", dir, page);

	i = _ifp_pop(dev, (ifp_u8 *)p + DOWNLOAD_PAGE_SIZE, DOWNLOAD_PAGE_SIZE);
	ifp_err_jump(i, out,
		"error downloading chuck #2 of (%#x, %#x)\n", dir, page);

out:
	return i;
}
int ifp_set_fat_page(struct ifp_device * dev, int dir, int page,
	void * p, int n)
{
	const int HEADER = 0x0010;
	const int DN_PAGE_SIZE = 0x0200;
	//ifp_u8 buf[DN_PAGE_SIZE + HEADER];
	ifp_u8 * buf = dev->b1;
	int i = 0;

	IFP_BUG_ON(n < DN_PAGE_SIZE);

	//Create the upload payload.
	memset(buf, 0, HEADER);
	memcpy(buf + HEADER, p, DN_PAGE_SIZE);
	((ifp_u16 *)(buf))[0] = ifp_os_cpu_to_le16(dir);
	((ifp_u16 *)(buf))[1] = ifp_os_cpu_to_le16(page);

	i = _ifp_set_buffer_size(dev, DN_PAGE_SIZE + HEADER, 1);
	ifp_err_jump(i, out, "error setting buffer size");

	i = _ifp_push(dev, buf, DN_PAGE_SIZE + HEADER);
	ifp_err_jump(i, out, "error pushing data for (%d, %d)",
		dir, page);

	i = _ifp_set_fat_page(dev);
	ifp_err_jump(i, out, "error setting FAT page (%d, %d)", dir, page);

out:
	return i;
}


